########################################################################################################################

__doc__ = \
    """
    This add the oddity fixing functionality. Generally odd corner cases.
    Formerly part of collapse_ring.py
    """

########################################################################################################################

from rdkit import Chem

from ._base import _RectifierBase


class _RectifierOdd(_RectifierBase):

    def prevent_oddities(self):
        self._prevent_allene()

    # ====== Private methods ===========================================================================================

    def _prevent_allene(self):
        for atom in self.rwmol.GetAtoms():
            if atom.GetAtomicNum() < 14:
                n = []
                for bond in atom.GetBonds():
                    if bond.GetBondType().name in ('DOUBLE', 'TRIPLE'):
                        n.append(bond)
                    else:
                        pass
                if len(n) > 2:
                    # this is a mess!
                    self.log.info(f'Allene issue: {n} double bonds on {atom.GetSymbol()} atom {atom.GetIdx()}!')
                    for bond in n:
                        bond.SetBondType(Chem.BondType().SINGLE)
                elif len(n) == 2:
                    # downgrade the higher bonded one!
                    others = [a for bond in n for a in (bond.GetBeginAtom(), bond.GetEndAtom()) if
                              a.GetIdx() != atom.GetIdx()]
                    others = sorted(others, key=lambda atom: sum([b.GetBondTypeAsDouble() for b in atom.GetBonds()]))
                    self.log.info(f'Allene removed between {atom.GetIdx()} and {[a.GetIdx() for a in others]}')
                    self.rwmol.GetBondBetweenAtoms(atom.GetIdx(), others[-1].GetIdx()).SetBondType(Chem.BondType.SINGLE)
                else:
                    pass
            else:
                continue
        self.modifications.append(self.mol)

    def _prevent_overclose(self):
        pass




from __future__ import annotations
########################################################################################################################

__doc__ = \
    """
Fix issue in auto-merging.
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "2021 A.D."
__license__ = "MIT"
__version__ = "0.1"
__citation__ = "See Fragmenstein"

########################################################################################################################

from ._base import _RectifierBase  # provides the __init__ and shared methods
from ._ring import _RectifierRing  # fixes rings
from ._odd import _RectifierOdd  # fixes specific oddities
from ._valence import _RectifierValence  # fixes valence
from ._overclose import _RectifierOverclose  # fixes extreme overlaps
from rdkit import Chem


class Rectifier(_RectifierOverclose, _RectifierRing, _RectifierOdd, _RectifierValence):
    """
    Fixes the nastiness.

    Do note that that the Chem.Mol does not get modified in place.
    ``.rwmol`` does and is a Chem.RWMol. The ``.mol`` is a Chem.Mol.

    The steps can be found in ``.modifications``.

    The .log log is not given a handler.

    New atoms with have the bool prop ``_Novel``.

    Does not link distant atoms. For that see joining methods in Monster.

    >>> Rectifier(mol).fix().mol
    """

    def fix(self) -> Rectifier:
        self.absorb_overclose()  # from _RectifierOverclose
        self.fix_rings()  # from _RectifierRing
        self.prevent_oddities()  # from _RectifierOdd
        self.ununspecified_bonds()  # from _RectifierValence
        self.triage_rings()  # from _RectifierValence
        Chem.Cleanup(self.rwmol)
        self.fix_issues()  # from _RectifierValence
        Chem.SanitizeMol(self.rwmol, sanitizeOps=Chem.SanitizeFlags.SANITIZE_ALL)
        return self



########################################################################################################################

__doc__ = \
    """
    This add the ring fixing functionality. Fusing etc.
    Formerly part of collapse_ring.py
    """

########################################################################################################################

import itertools
from collections import Counter
from typing import Optional, Tuple, Dict, Union, List, Set

from rdkit import Chem
from rdkit.Geometry.rdGeometry import Point3D

from ._base import _RectifierBase


# ======================================================================================================================


class _RectifierRing(_RectifierBase):

    def fix_rings(self):
        self._prevent_bonded_to_bridgeheads()
        self._prevent_conjoined_ring()
        self._prevent_weird_rings()

    def _prevent_bonded_to_bridgeheads(self):
        """
        A 3-aromoatic-bond carbon cannot be bonded to anything else (except if a spiro)
        It removes the offending bond but does nothig more...
        """
        # list of dict shared, ring_atom_A_idxs, ring_atom_B_idxs
        bridge_infos = self._get_bridges()
        for info in bridge_infos:
            for idx in info['shared']:
                atom: Chem.Atom = self.rwmol.GetAtomWithIdx(idx)
                # middle atom:
                if len(atom.GetNeighbors()) <= 2:
                    continue
                # heavy-heavy atom:
                elif atom.GetAtomicNum() > 8:
                    continue
                # aromatic atom:
                elif atom.GetIsAromatic() or \
                   any([bond.GetBondType() == Chem.BondType.AROMATIC for bond in atom.GetBonds()]):
                    neigh_idxs: Set[int] = {neigh.GetIdx() for neigh in atom.GetNeighbors()}
                    other_idxs = neigh_idxs.difference(info['ring_atom_A_idxs'])\
                                           .difference(info['ring_atom_B_idxs'])
                    for idx in other_idxs:
                        self.log.info('non ring atom bonded to bridgehead')
                        self.rwmol.RemoveBond(atom.GetIdx(), idx)


    def _prevent_conjoined_ring(self) -> None:
        """
        This kills bridging bonds with not atoms in the bridge within rings.
        So it is bridged, fused and spiro safe.
        It removes only one bond, so andamantane/norbornane are safe.
        """
        c = Counter([i for ring in self._get_ring_info() for i in ring])
        nested = [k for k in c if c[k] >= 3]
        pairs = [(idx_a, idx_b) for idx_a, idx_b in itertools.combinations(nested, r=2) if
                 self.rwmol.GetBondBetweenAtoms(idx_a, idx_b) is not None]
        rank = sorted(pairs, key=lambda x: c[x[0]] + c[x[1]], reverse=True)
        if len(rank) > 0:
            idx_a, idx_b = rank[0]
            self.rwmol.RemoveBond(idx_a, idx_b)  # SetBoolProp('_IsRingBond') is not important
            self.log.info(f'Zero-atom bridged ring issue: bond between {idx_a}-{idx_b} removed')
            # re-run:
            self._prevent_conjoined_ring()
        self.modifications.append(self.mol)  # not self.rwmol as I want a snapshot

    def _get_bridges(self) -> List[Dict[str, Union[Tuple[int]]]]:
        """
        The output includes not only the bridgeheads but the middle ones too
        list of dict shared, ring_atom_A_idxs, ring_atom_A_idxs
        """
        bridge_info = []
        ringatoms = self._get_ring_info()  # GetRingInfo().AtomRings()
        for ring_A, ring_B in itertools.combinations(ringatoms, r=2):
            shared = set(ring_A).intersection(set(ring_B))
            if len(shared) == 0:
                continue
            bridge_info.append(dict(shared=shared,
                                    ring_atom_A_idxs=ring_A,
                                    ring_atom_B_idxs=ring_B)
                                   )
        return bridge_info


    def _prevent_weird_rings(self, iteration=0):
        bridge_info: List[Dict[str, Tuple[int]]] = self._get_bridges()
        for info in bridge_info:
            ring_A: Tuple[int] = info['ring_atom_A_idxs']
            ring_B: Tuple[int] = info['ring_atom_B_idxs']
            shared: Tuple[int] = info['shared']
            if len(shared) == 0:
                # no longer accessible
                self.log.debug(f'This molecule ({self.name}) has some separate rings')
                pass  # separate rings
            elif len(shared) < self.atoms_in_bridge_cutoff and \
                    self.atoms_in_bridge_cutoff >= 2 \
                    and len(ring_A) == len(ring_B):
                # adamantene/norbornane/tropinone kind of thing
                self.log.warning(f'This molecule ({self.name}) has a bridge: leaving')
                pass  # ideally check if planar...
            elif len(shared) == 1:
                self.log.debug(f'This molecule ({self.name}) has a spiro bicycle')
                pass  # spiro ring.
            elif len(shared) == 2:
                self.log.debug(f'This molecule ({self.name}) has a fused ring')
                if self.rwmol.GetBondBetweenAtoms(*shared) is not None:
                    pass  # indole/naphtalene
                    small, big = sorted([ring_A, ring_B], key=lambda ring: len(ring))
                    if len(small) == 4:
                        self.log.warning(f'This molecule ({self.name}) ' +
                                             'has a benzo-azetine–kind-of-thing: expanding to indole')
                        # Chem.MolFromSmiles('C12CCCCC1CC2')
                        # benzo-azetine is likely an error: add and extra atom
                        a, b = set(small).difference(big)
                        self._place_between(a, b)
                    elif len(small) == 3:
                        self.log.warning(f'This molecule ({self.name}) '+
                                             'has a benzo-cyclopropane–kind-of-thing: expanding to indole')
                        # Chem.MolFromSmiles('C12CCCCC1C2')
                        # benzo-cyclopronane is actually impossible at this stage.
                        a = list(set(small).difference(big))[0]
                        for b in shared:
                            self._place_between(a, b)
                    else:
                        pass  # indole and nathalene
                elif (len(ring_A), len(ring_B)) == (6, 6):
                    raise Exception('This is utterly impossible')
                else:
                    self.log.warning(f'mysterious ring system {len(ring_A)} + {len(ring_B)}')
                    pass  # ????
            elif len(shared) < self.atoms_in_bridge_cutoff:
                # adamantene/norbornane/tropinone kind of thing
                self.log.warning(f'This molecule ({self.name}) has a bridge: leaving')
                pass  # ideally check if planar...
            else:
                self.log.warning(f'This molecule ({self.name}) has a bridge that will be removed')
                self._prevent_bridge_ring(ring_A)
                # start from scratch.
                if iteration < 3:
                    self._prevent_weird_rings(iteration=iteration+1)
                else:
                    self.log.warning(f'Too many trials to remove bridge.')
        self.modifications.append(self.mol)  # not self.rwmol as I want a snapshot

    # ===== Dependencies of prevent weird rings =================================================================================

    def _place_between(self, a: int, b: int, aromatic: Optional[bool] = None, atomic_number: int = 6) -> None:
        """
        Places an C atom, possibly of type aromatic, between atom of index a, and of b.

        :param a: index of atom A
        :param b: index of atom B
        :param aromatic: bool of aromaticity (False = Single, None = copy, True = aromatic)
        :param atomic_number: Carbon is 6.
        :return:
        """
        oribond = self.rwmol.GetBondBetweenAtoms(a, b)
        if oribond is None:
            self.log.critical(f'FAIL. There should be a bond btween {a} and {b}')
            return None  # fail
        elif aromatic is True:
            bt = Chem.BondType.AROMATIC
        elif aromatic is False:
            bt = Chem.BondType.SINGLE
        else:
            bt = oribond.GetBondType()
        idx = self.rwmol.AddAtom(Chem.Atom(atomic_number))
        neoatom = self.rwmol.GetAtomWithIdx(idx)
        atom_a = self.rwmol.GetAtomWithIdx(a)
        atom_b = self.rwmol.GetAtomWithIdx(b)
        if aromatic:
            neoatom.SetIsAromatic(True)
            atom_a.SetIsAromatic(True)
            atom_b.SetIsAromatic(True)
        # prevent constraints
        neoatom.SetBoolProp('_Novel', True)
        atom_a.SetBoolProp('_Novel', True)
        atom_b.SetBoolProp('_Novel', True)
        # fix position
        conf = self.rwmol.GetConformer()
        pos_A = conf.GetAtomPosition(a)
        pos_B = conf.GetAtomPosition(b)
        x = pos_A.x / 2 + pos_B.x / 2
        y = pos_A.y / 2 + pos_B.y / 2
        z = pos_A.z / 2 + pos_B.z / 2
        conf.SetAtomPosition(idx, Point3D(x, y, z))
        # fix bonds
        self.rwmol.RemoveBond(a, b)
        self.rwmol.AddBond(a, idx, bt)
        self.rwmol.AddBond(b, idx, bt)

    def _prevent_bridge_ring(self, examplar: Tuple[int]) -> None:
        # examplar is ring
        ringatoms = self._get_ring_info()  # GetRingInfo().AtomRings()
        ringatoms = [ring for ring in ringatoms if set(ring).intersection(examplar)]
        ring_idx = list(range(len(ringatoms)))
        shared_count = {}
        for ra, rb in itertools.combinations(ring_idx, r=2):
            shared_count[(ra, rb)] = len(set(ringatoms[ra]).intersection(set(ringatoms[rb])))
        if len(shared_count) == 0:
            return None
        ra, rb = list(shared_count.keys())[0]
        shared = list(set(ringatoms[ra]).intersection(ringatoms[rb]))
        has_bond = lambda a, b: self.rwmol.GetBondBetweenAtoms(a, b) is not None
        pairs = [(a, b) for a, b in itertools.combinations(shared, r=2) if has_bond(a, b)]
        c = Counter([i for pair in pairs for i in pair])
        ring_A, ring_B = ringatoms[ra], ringatoms[rb]
        small, big = sorted([ring_A, ring_B], key=lambda ring: len(ring))
        inners = [i for i in c if c[i] > 1]
        x = list(set(shared).difference(inners))
        if len(x) != 2:
            self.log.critical(
                f'This is impossible. {ringatoms} share {shared} with {inners} in the inside and {x} on the edge?')
            return None
        a, b = x
        if len(big) > 6:
            self.log.warning(f'Removing {len(inners)} bridging atoms and replacing with fused ring')
            # bond the vertices
            bt = Chem.BondType.SINGLE  # ???
            if self.rwmol.GetBondBetweenAtoms(a, b) is None:
                self.rwmol.AddBond(a, b, bt)
            else:
                self.log.warning('This is really odd! Why is there a bond already??')
            # remove the middle atoms.
            for i in sorted(inners, reverse=True):
                self.rwmol.RemoveAtom(i)
        else:
            self.log.warning(f'Shriking the smaller ring to change from bridged to fused.')
            # get the neighbour in the small atom to a vertex.
            neighs = [neigh for neigh in self.rwmol.GetAtomWithIdx(a).GetNeighbors() if
                      neigh.GetIdx() not in shared and neigh.GetIdx() in small]
            neigh = sorted(neighs, key=lambda atom: atom.GetSymbol() != 'C')[0]
            bt = self.rwmol.GetBondBetweenAtoms(a, neigh.GetIdx()).GetBondType()
            self.rwmol.RemoveBond(a, neigh.GetIdx())
            new_neigh = [neigh for neigh in self.rwmol.GetAtomWithIdx(a).GetNeighbors() if neigh.GetIdx() in shared][0]
            self.rwmol.AddBond(neigh.GetIdx(), new_neigh.GetIdx(), bt)
            neigh.SetBoolProp('_Novel', True)
            new_neigh.SetBoolProp('_Novel', True)
            self.rwmol.GetAtomWithIdx(a).SetBoolProp('_Novel', True)





from ._base import _RectifierBase
import numpy as np
from rdkit import Chem

from typing import List, Union
from collections import Counter

# ======================================================================================================================


class _RectifierOverclose(_RectifierBase):
    """
    Fragmenstein monster is better equipped to deal with overclose atoms
    This simply absorbs all which are egrigiously too close
    It is an extreme fallback
    """
    def absorb_overclose(self, distance_cutoff: float=0.6):
        """
        Never say never... This really should have an effect as Fragmenstein monster is build for this.
        But stuff always goes wrong
        """
        # make matrix
        distance_matrix: np.ndarray = Chem.Get3DDistanceMatrix(self.rwmol)
        # nan fill the self values
        np.fill_diagonal(distance_matrix, np.nan)
        # nan fill the upper triangle
        distance_matrix[np.triu_indices(distance_matrix.shape[0],)] = np.nan
        # get overclose indices as a list of two indices
        mergers: List[List[int]] = np.transpose( (distance_matrix < distance_cutoff).nonzero() ).tolist()  # noqa
        if len(mergers) == 0:
            return
        # sort the mergers by most common atoms
        commonality = dict(Counter(np.array(mergers).flatten()).most_common())
        mergers = sorted(mergers, key=lambda idxs: commonality[idxs[0]]+commonality[idxs[1]])
        # sort each pair by most common (and then bonded) first
        sort_indices = lambda i: self._get_atom_valence(self.rwmol.GetAtomWithIdx(i)) + 10 * commonality[i]
        mergers = [sorted(idxs, key=sort_indices, reverse=True) for idxs in mergers]
        self.log.critical(f'Absorption required for the atom pairs: {mergers}')
        # mark for deletion
        for _, i in mergers:
            self.rwmol.GetAtomWithIdx(i).SetBoolProp('DELETE', True)
        # copy bonds
        for keep_i, del_i in mergers:
            for neigh in self.rwmol.GetAtomWithIdx(del_i).GetNeighbors():  #: Chem.Atom
                neigh_i: int = neigh.GetIdx()
                if neigh_i == keep_i:
                    continue  # I expect keep is del's neigh
                del_bondtype: Chem.BondType =  self.rwmol.GetBondBetweenAtoms(neigh_i, del_i).GetBondType()
                previous_bond: Union[None, Chem.Bond] = self.rwmol.GetBondBetweenAtoms(keep_i, neigh_i)
                if not previous_bond:
                    self.rwmol.AddBond(keep_i, neigh_i, del_bondtype)
        # delete atoms
        while True:
            for atom in self.rwmol.GetAtoms():
                if atom.HasProp('DELETE'):
                    self.rwmol.RemoveAtom(atom.GetIdx())
                    break
            else:
                break




from setuptools import setup, find_packages
from warnings import warn
from importlib import util
import sys, os

if sys.version_info.major != 3 or sys.version_info.minor < 6:
    print(sys.version_info)
    raise SystemError('Module written for Python 3.6+.')

# ---------- Non pip modules  ------------------------------------------------------------------------------------------

if not util.find_spec('rdkit'):
    warn('This 3.6+ script **requires** rdkit which cannot be pip installed.' +
                              ' To install try either ' +
                              'conda install -c conda-forge rdkit or ' +
                              'sudo apt-get/brew install python3-rdkit or visit rdkit documentation.')

# ------------ description ---------------------------------------------------------------------------------------------

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    __doc__ = f.read()
descr = __doc__.split('\n')[1] # non-title line.

# ------------ Setup ---------------------------------------------------------------------------------------------------

setup(
    name='molecular_rectifier',
    version='0.1.3',
    python_requires='>3.6',
    packages=find_packages(),
    url='https://github.com/matteoferla/molecular_rectifier',
    license='MIT',
    author='Matteo Ferla',
    author_email='matteo.ferla@gmail.com',
    description=descr,
    long_description=__doc__,
    long_description_content_type='text/markdown',
)



########################################################################################################################

__doc__ = \
    """
    This add the oddity fixing functionality. Generally odd corner cases.
    Formerly part of collapse_ring.py
    """



########################################################################################################################
from rdkit import Chem
import itertools
from typing import List, Tuple, Union
from ._base import _RectifierBase


# when hits are combined they can result in odd valence and other issues.

class _RectifierValence(_RectifierBase):
    """
        Checks whether the valence is right and corrects by either shifting the element or adding a charge.
        With the following exceptions:

        * Texas carbon -> Sulfur
        * Hydrogen -> Fluoride shifted downwards
        * Carbon in aromatic ring -> single bond ring
    """
    # ========= Three main steps =======================================================================================

    def ununspecified_bonds(self) -> None:
        """
        This get run during instantiation. It is the first one run.
        BondType.UNSPECIFIED ==> BondType.SINGLE

        :return: None
        """
        for bond in self.rwmol.GetBonds():
            if bond.GetBondType().name == 'UNSPECIFIED':
                self.log.debug(f'Fixing unspecified bond {bond.GetIdx()}')
                bond.SetBondType(Chem.BondType.SINGLE)
                # debug:
                self.modifications.append(self.mol)

    def triage_rings(self) -> None:
        """
        This get run during instantiation. It is the second one run.
        It is also run again by `fix_issues` the third step after a weird valence change.
        Deals with rings.

        :return: None
        """
        # upgrade to aromatic if aromatic.
        rings = self._get_ring_info()
        for ring in rings:
            if self._is_aromatic_ring(ring, rings):
                for i, ni in self._get_ring_neighbors(ring):
                    self.rwmol.GetBondBetweenAtoms(i, ni).SetBondType(Chem.BondType.AROMATIC)
        # downgrade to single if non-ring aromatic
        for i, atom in enumerate(self.rwmol.GetAtoms()):
            if any([i in r for r in rings]):
                continue  # ring
            else:  # non-ring
                for bond in atom.GetBonds():
                    if bond.GetBondType().name == 'AROMATIC':
                        self.log.debug(f'downgrading bond {i}')
                        bond.SetBondType(Chem.BondType.SINGLE)
        # aromatics
        for ring in sorted(rings, key=self._is_aromatic_ring):
            if self._is_aromatic_ring(ring, rings):  # the nonaromatic rings will be done first.
                for i in ring:
                    self.rwmol.GetAtomWithIdx(i).SetIsAromatic(True)
            else:
                for i in ring:
                    self.rwmol.GetAtomWithIdx(i).SetIsAromatic(False)
        self.modifications.append(self.mol)  # may not have changed.

    def fix_issues(self, _previous=None) -> None:
        """
        This get run during instantiation. It is the third and final one run before sanitization.
        Deals with a variety of problems.
        It calls itself until no problems according to `Chem.DetectChemistryProblems` exits.
        It is a bit shoddy and any oddity likely steps from here. TODO

        :return: None
        """
        self.modifications.append(self.mol)  # may not have changed.
        problems = Chem.DetectChemistryProblems(self.rwmol)
        if self._iterations_done > 100:
            self.log.error(f'Iterations maxed out!')
            return None
        elif self._subiterations_done > 5:
            self.log.error(f'Unfixable')
            return None
        elif len(problems) == 0:
            return None
        else:
            self.log.debug(f'(Iteration: {self._iterations_done}) N problems {len(problems)}')
            p = problems[0]
            self.log.debug(f'(Iteration: {self._iterations_done}) Issue {p.GetType()}: {p.Message()}')
            if p.Message() == _previous:
                self.triage_rings()
            ############################################################
            if p.GetType() == 'KekulizeException':
                if p.Message() != _previous:
                    N = self._get_nitrogens(p.GetAtomIndices())
                    if len(N) > 0 and self._nitrogen_protonate(N, p.Message()):
                        pass  # been fixed.
                    else:
                        # triage rings should have altered any not ring atoms that are aromatic.
                        # self._get_ring_info()
                        # so it is likely a hetatom thing.
                        self.log.info(f'Ring triages seems to have failed. Is it a valence thing?')
                        valence_issues = [self._has_correct_valence(i) for i in p.GetAtomIndices()]
                        if not all(valence_issues):
                            for i in p.GetAtomIndices():
                                self.fix_valence(i)
                        else:
                            self.log.warning(f'Attempting default valency (not max)')
                            self._valence_mode = 'default'
                            for i in p.GetAtomIndices():
                                self.fix_valence(i)
                            self._valence_mode = 'max'
                else:
                    for i in p.GetAtomIndices():
                        self.downgrade_ring(self.rwmol.GetAtomWithIdx(i))
                    self.triage_rings()
            ############################################################
            elif p.GetType() == 'AtomKekulizeException' and 'non-ring atom' in p.Message():
                atom = self.rwmol.GetAtomWithIdx(p.GetAtomIdx())
                atom.SetIsAromatic(False)
                self.log.debug(f'Atom {p.GetAtomIdx()} set to non-aromatic.')
                for bond in atom.GetBonds():
                    bond.SetBondType(Chem.BondType.SINGLE)
            elif p.GetType() == 'AtomKekulizeException' and 'Aromatic bonds on non aromatic atom' in p.Message():
                atom = self.rwmol.GetAtomWithIdx(p.GetAtomIdx())
                self.log.debug(f'Atom {p.GetAtomIdx()} set to aromatic.')
                atom.SetIsAromatic(True)
            ############################################################
            elif p.GetType() == 'AtomValenceException':
                i = p.GetAtomIdx()
                self.fix_valence(i)
            else:
                self.log.error('???', p.GetType(), p.Message())
            self._iterations_done += 1
            if _previous != p.Message():
                self.log.debug(f'{self._iterations_done} appears successful.')
                self._subiterations_done = 0
            else:
                self._subiterations_done += 1
                self.log.debug(f'{self._iterations_done} appears unsuccessful.')
            return self.fix_issues(_previous=p.Message())

    # ========= Methods that circumvent the nonsanitization ============================================================

    def _get_valence_difference(self, atom: Chem.Atom) -> int:
        pt = Chem.GetPeriodicTable()
        valence = self._get_atom_valence(atom)
        if self._valence_mode == 'max':
            maxv = max(pt.GetValenceList(atom.GetAtomicNum()))
            return valence - maxv
        else:
            d = pt.GetDefaultValence(atom.GetAtomicNum())
            return valence - d

    def _has_correct_valence(self, atom: Union[Chem.Atom, int]):
        if isinstance(atom, Chem.Atom):
            return self._get_valence_difference(atom) <= 0
        elif isinstance(atom, int):
            atom = self.rwmol.GetAtomWithIdx(atom)
            return self._get_valence_difference(atom) <= 0

    # ========= rings ==================================================================================================

    def _get_atoms_at_fusion(self, ring, rings):
        fused = []
        if rings is not None:
            for other in rings:
                if other == ring:
                    pass
                elif not set(ring).isdisjoint(other):
                    fused.extend(set(ring).intersection(other))
        return fused

    def _is_aromatic_ring(self, ring: Tuple[int], rings=None) -> bool:
        """
        :param ring: GetRingInfo().AtomRings() entry
        :return:
        """
        fused = self._get_atoms_at_fusion(ring, rings)
        for i in list(set(ring).difference(fused)):
            atom_i = self.rwmol.GetAtomWithIdx(i)
            for n in atom_i.GetNeighbors():
                ni = n.GetIdx()
                if ni in ring:
                    if self.rwmol.GetBondBetweenAtoms(i, ni).GetBondType().name == 'AROMATIC':
                        return True
        else:
            return False

    def _get_ring_neighbors(self, ring: Tuple[int]) -> List[Tuple[int, int]]:
        """
        :param ring: GetRingInfo().AtomRings() entry
        :return: list of pairs of indices that are neighbors in the ring
        """
        rns = []
        for i in ring:
            atom = self.rwmol.GetAtomWithIdx(i)
            for n in atom.GetNeighbors():
                ni = n.GetIdx()
                if ni in ring:
                    rns.append((i, ni))
        return rns

    def _get_aroma(self, atom, this_bond):
        # determine if the bond of the atom is aromatic
        return [b for b in atom.GetBonds() if b.GetIdx() != this_bond and b.GetBondType().name == 'AROMATIC']

    def _get_other(self, bond, these_atoms):
        others = [a for a in (bond.GetBeginAtom(), bond.GetEndAtom()) if a.GetIdx() not in these_atoms]
        if others:
            other = others[0]
            other.SetIsAromatic(False)
            return other

    def downgrade_ring(self, atom: Chem.Atom):
        ## very crappy way of doing this
        self.log.debug(f'downgrading whole ring!')
        atom.SetIsAromatic(False)
        ringinfo = self._get_ring_info(mode='atom')
        get_atomrings = lambda ai: [ring for ring in ringinfo if ai in ring]
        atomrings = get_atomrings(atom.GetIdx())
        for atomring in atomrings:
            rnieghs = self._get_ring_neighbors(atomring)
            for n1, n2 in rnieghs:
                self.rwmol.GetAtomWithIdx(n1).SetIsAromatic(False)
                self.rwmol.GetAtomWithIdx(n2).SetIsAromatic(False)
                self.rwmol.GetBondBetweenAtoms(n1, n2).SetBondType(Chem.BondType.SINGLE)
        for atomring in atomrings:
            rnieghs = self._get_ring_neighbors(atomring)
            for n1, n2 in rnieghs:
                if self._get_valence_difference(self.rwmol.GetAtomWithIdx(n1)) <= -2 and \
                        self._get_valence_difference(self.rwmol.GetAtomWithIdx(n2)) <= -2:
                    self.rwmol.GetBondBetweenAtoms(n1, n2).SetBondType(Chem.BondType.DOUBLE)

        # if len(self._get_rings(atom.GetIdx())) == 1:
        #     for bond in atom.GetBonds():
        #         bond.SetBondType(Chem.BondType.SINGLE)
        #         other = self._get_other(bond, [atom.GetIdx()])
        #         aro = self._get_aroma(other, bond.GetIdx())
        #         if aro:
        #             aro[0].SetBondType(Chem.BondType.DOUBLE)
        #             doubleother = self._get_other(aro[0], [atom.GetIdx(), other.GetIdx()])
        #             for b in doubleother.GetBonds():
        #                 if b.GetBondType() == Chem.BondType.AROMATIC:
        #                     b.SetBondType(Chem.BondType.SINGLE)
        #                 neigh = self._get_other(b, [doubleother.GetIdx()])
        #                 if neigh:
        #                     neigh.SetIsAromatic(False)

    # ========= Sanitization based fixes ===============================================================================

    def _nitrogen_protonate(self, nitrogens, previous):
        """

        :param nitrogens: list of Nitrogens
        :param previous:
        :return:
        """
        def reset():
            for n in nitrogens:
                self.rwmol.GetAtomWithIdx(n).SetNumExplicitHs(0)

        reset()
        p = Chem.DetectChemistryProblems(self.rwmol)
        if len(p) == 0 or p[0].Message() != previous:
            return True
        for i in range(1, len(nitrogens)):
            for c in itertools.combinations(nitrogens, i):
                reset()
                for n in c:
                    self.rwmol.GetAtomWithIdx(n).SetNumExplicitHs(1)
                p = Chem.DetectChemistryProblems(self.rwmol)
                if len(p) == 0 or p[0].Message() != previous:
                    return True
        return False

        self.log.debug(f'KekulizeException likely caused by nitrogen')

    # ========= other helpers ==========================================================================================

    def _get_nitrogens(self, indices):
        """
        Called when ``KekulizeException`` happends during ``.fix_issues``

        :param indices:
        :return:
        """
        return [i for i in indices if self.rwmol.GetAtomWithIdx(i).GetSymbol() == 'N']

    # ========= shift/charge ===========================================================================================

    def _adjust_for_fix_valence(self, atom):
        df = self._get_valence_difference(atom)
        ori = atom.GetSymbol()
        if self.valence_correction == 'charge':
            atom.SetFormalCharge(df)
        elif self.valence_correction == 'element':
            # ## correct row
            n = atom.GetAtomicNum()
            if n == 1:
                atom.SetAtomicNum(8)
            elif n > 10:
                n = (n % 8) - 2 + 8
                atom.SetAtomicNum(n)
            # ## correct column
            if len(atom.GetNeighbors()) > 4:
                self._break_bonds(atom)
            # elif len(atom.GetNeighbors()) > 4 and n <= 16: # S...
            #     atom.SetAtomicNum(16)
            elif n - df < 6:  # C -> B no!
                for bond in atom.GetBonds():
                    bond.SetBondType(Chem.BondType.SINGLE)
            else:  # N, O, F etc.
                atom.SetAtomicNum(int(n - df))
            self.log.info(f'Shifting atom from {ori} to {atom.GetSymbol()}')
        else:
            raise ValueError(f'self.valence_correction can only be "element"/"charge" not {self.valence_correction}.')

    def _break_bonds(self, atom):
        """
        Extreme last ditch. Breaks off all non-ring bonds to atom.
        Will likely trigger emergency_joining.

        :param atom:
        :return:
        """
        self.log.warning(f'In molecule ({self.name}) breaking bond to atom {atom.GetIdx()}')
        ring_indices = [a for ring in self._get_ring_info() for a in ring]
        for neigh in atom.GetNeighbors():
            if neigh.GetIdx() in ring_indices or neigh.GetSymbol() == '*':
                continue
            else:
                self.rwmol.RemoveBond(atom.GetIdx(), neigh.GetIdx())


    def fix_valence(self, i):
        atom = self.rwmol.GetAtomWithIdx(i)
        atom.SetFormalCharge(0)
        atom.SetNumExplicitHs(0)
        self.log.debug(f'{i} {atom.GetSymbol()}: {len(atom.GetNeighbors())} bonds {self._get_atom_valence(atom)}')
        if self._has_correct_valence(atom):
            self.log.debug('\tValence seems correct')
            return None
        elif atom.GetSymbol() == 'C' and len(atom.GetNeighbors()) > 4:
            self.log.debug('\ttexas carbon --> S')
            atom.SetAtomicNum(16)
        elif atom.GetSymbol() == 'C' and atom.GetIsAromatic() and len(atom.GetNeighbors()) == 4:
            self.log.debug('\tDowngrading ring')
            self.downgrade_ring(atom)
        elif atom.GetSymbol() == 'C':
            for bond in atom.GetBonds():
                bond.SetBondType(Chem.BondType.SINGLE)
        else:
            self._adjust_for_fix_valence(atom)
        # did it work?
        if self._has_correct_valence(atom):
            return self.rwmol
        else:
            return self.fix_valence(i)



########################################################################################################################

__doc__ = \
    """
    This has the methods common to ring and odd.
    Formerly part of collapse_ring.py
    """

########################################################################################################################

import logging
from typing import Tuple

from rdkit import Chem


class _RectifierBase:
    log = logging.getLogger(__name__)

    def __init__(self, mol: Chem.Mol,
                 atoms_in_bridge_cutoff: int = 2,
                 valence_correction: str = 'element',
                 debug: bool = False):
        """
        Instantiates but does not call ``fix`` (or its specific methods).

        :param mol: Does not get edited. ``.mol`` does (but is a``Chem.RWMol``, so use ``mol.GetMol()``)
        # how many bridge atoms can be deleted? (0 = preserves norbornane, 1 = preserves monsterantane)

        :param valence_correction:
        :param debug:
        """
        self.debug = bool(debug)
        if valence_correction not in ('charge', 'element'):
            raise ValueError(f'valence_correction "{valence_correction} id not charge/element')
        self.valence_correction = str(valence_correction)
        self.atoms_in_bridge_cutoff = int(atoms_in_bridge_cutoff)
        self.rwmol = Chem.RWMol(mol)
        self.modifications = []  # keeping track of steps
        self._valence_mode = 'max'
        self._iterations_done = 0
        self._subiterations_done = 0

    @property
    def mol(self) -> Chem.Mol:
        return self.rwmol.GetMol()

    @property
    def name(self) -> str:
        if self.rwmol.HasProp("_Name"):
            return self.rwmol.GetProp("_Name")
        else:
            return 'compound'

    def _get_ring_info(self, mode='atom') -> Tuple[Tuple[int]]:
        """
        you cannot get ring info on an unsanitized mol. Ironically I need ring info for sanitization

        :param mode: bond|atom
        :return: same as mol.GetRingInfo().AtomRings() or .BondRings()
        """
        mol2 = Chem.Mol(self.mol)
        for bond in mol2.GetBonds():
            bond.SetBondType(Chem.BondType.UNSPECIFIED)
        for atom in mol2.GetAtoms():
            atom.SetIsAromatic(False)
            atom.SetAtomicNum(0)
        Chem.SanitizeMol(mol2)
        if mode == 'atom':
            return mol2.GetRingInfo().AtomRings()
        elif mode == 'bond':
            return mol2.GetRingInfo().BondRings()
        else:
            raise ValueError(f'Unknown mode {mode}')


    def _get_atom_valence(self, atom: Chem.Atom):
        """
        Cannot get the normal way as it cannot be sanitised.

        :param atom:
        :return:
        """
        valence = 0
        for bond in atom.GetBonds():
            valence += bond.GetBondTypeAsDouble()
        return valence - atom.GetFormalCharge()
